package com.adafruit.bluefruit.le.connect.app;

import android.app.Activity;
import android.app.Application;
import android.content.Context;
import android.widget.Toast;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Scanner;

/**
 *
 * @author Dan Carroll
 */

class UartDataStorage extends Application {

    private String buffer;

    private File myExternalFile = null;
    private Context context;

    private BufferedWriter bw;
    private FileOutputStream fos;
    private Scanner sc;
    String filename;

    UartDataStorage() {
        strings = new String[27000];
        strings[0] = "";
        buffer = "";
        uart_crc = 0;
    }

    void setContext(Context c)
    { context = c;}

    void updateFile()
    {
        try {
            if (myExternalFile == null) {
                myExternalFile = new File(context.getExternalFilesDir("MyFileStorage"), filename + ".txt");
                bw = new BufferedWriter(new FileWriter(myExternalFile, true));
                buffer = "";
                sc = new Scanner(buffer);
            } else {

                myExternalFile = null;
                bw.flush();
                bw.close();
                sc.close();
            }
        }
        catch (IOException e) {
                e.printStackTrace();
        }
    }

    String[] strings;
    String temp;
    int dataIndex = 0;
    int measuredLength = 0;

    /**
     * Assumes text is a portion of a long string of lines coming in as input to be saved to filename
     *
     * @param text one packet of text which is the input ot be saved.
     * @param activity the app's activity, used to create Toast messages.
     * @return 0 if the input has reached its end and no errors were detected, -1 if an error was found, 1 if the input is not at the end.
     */
    int saveToInternalStorage(String text, Activity activity)
    {
        buffer = buffer + text;
        temp = "";

        if(buffer.contains("\n")) {
            char c = '\n';

            temp = buffer.substring(0, buffer.indexOf(c));

            if(!temp.contains("CRC") && !temp.contains("Log")) // Only perform CRC calculations on the data lines.
                measuredLength += temp.length();

            if(temp.length() != 0) {
                strings[dataIndex++] = temp;
            }
            buffer = buffer.substring(buffer.indexOf(c) + 1);
        }

        if(temp.equals("Log END--------------------")) {
            String transmittedCRC = strings[dataIndex-2].substring(5, 9);

            //int correctLength = 24 * Integer.parseInt(strings[1].substring(11, strings[1].indexOf("/")));

            int correctLength = measuredLength-1; // just for testing

            if(correctLength == measuredLength-1) {
                Toast.makeText(activity, "Successfully downloaded all data from the board. Saving data to internal storage. ", Toast.LENGTH_SHORT).show();

                for (int z = 3; z < dataIndex - 2; z++) {
                    String formattedString = "";

                    formattedString += strings[z].substring(0, 2) + "/" + strings[z].substring(2, 4) + "/" + strings[z].substring(4, 6)
                                + " " + strings[z].substring(6, 8) + ":" + strings[z].substring(8, 10) + ":" + strings[z].substring(10, 12)
                                + " " + strings[z].substring(12, 22) + " " + strings[z].substring(22, 26);

                    save(formattedString + "\n");
                }

                Toast.makeText(activity, "Successfully saved all the data to " + filename + ".txt", Toast.LENGTH_LONG).show();

                updateFile();

                uart_crc = 0;
                measuredLength = 0;
                dataIndex = 0;
                buffer = "";
                return 0;
            }
            else {
                Toast.makeText(activity, "Transmission error. Data not saved.", Toast.LENGTH_LONG).show();

                uart_crc = 0;
                measuredLength = 0;
                dataIndex = 0;
                buffer = "";
                return -1;
            }

        }

        return 1;
    }

    private void save(String text)
    {
        try {
            bw.write(text);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    short uart_crc;
    final short CRC16_POLY = 0x1021;

    private void do_crc(char c)
    {
        char j;

        for (j = 0; j < 8; j++)
        {
            if (((c >> 7) ^ (uart_crc >> 15)) == 1)
            {
                uart_crc <<= 1;
                uart_crc ^= CRC16_POLY;
            }
            else
            {
                uart_crc <<= 1;
            }
            c <<= 1;
        }
    }

}
